#!/usr/bin/env python3
from pathlib import Path
from typing import List

from yt_dlp import YoutubeDL


class YoutubePlaylistHandler:
    def __init__(self, cookies_file: str = None):
        if cookies_file is not None:
            cookies_file = Path(str(cookies_file)).expanduser().resolve()
            if not cookies_file.exists():
                raise ValueError(
                    f'Youtube cookies file "{cookies_file}" does not exist. '
                    'Please check your catt config file.'
                )
        self.ydl = YoutubeDL(
            {
                'cookiefile': str(cookies_file),
                'skip_download': True,
                'quiet': True,
                'ignoreerrors': True,
                'no_warnings': True,
                'extract_flat': 'in_playlist',
            }
        )

    @staticmethod
    def _get_ids_from_ytdl_json(ytdl_json: dict) -> List[str]:
        return [item['id'] for item in ytdl_json['entries']]

    def resolve_playlist(self, s: str) -> List[str]:
        """
        Resolve a playlist into a list of video URLs.
        """
        try:
            ids = self._get_ids_from_ytdl_json(self.ydl.extract_info(s, download=False))
            return [f'https://www.youtube.com/watch?v={id_}' for id_ in ids]
        except TypeError:
            return []



from typing import Dict

from textual.app import App
from textual.containers import Container
from textual.screen import Screen

from lolcatt.casting.caster import Caster
from lolcatt.ui.lolcatt_controls import LolCattControls
from lolcatt.ui.lolcatt_device_info import LolCattDeviceInfo
from lolcatt.ui.lolcatt_playback_info import LolCattPlaybackInfo
from lolcatt.ui.lolcatt_progress import LolCattProgress
from lolcatt.ui.lolcatt_url_input import LolCattUrlInput


class RemoteScreen(Screen):
    """A screen for the remote control UI."""

    def compose(self):
        yield Container(
            LolCattDeviceInfo(),
            LolCattPlaybackInfo(),
            LolCattProgress(),
            LolCattControls(),
            LolCattUrlInput(),
            id='app',
        )


class LolCatt(App):
    """The main application class for lolcatt."""

    CSS_PATH = 'ui/lolcatt.css'

    def __init__(self, device_name: str = None, config: Dict = None, *args, **kwargs):
        self.config = config
        self.caster = Caster(device_name, config=config)
        self.remote_screen = None
        super().__init__(*args, **kwargs)

    def on_mount(self):
        self.remote_screen = RemoteScreen()
        self.install_screen(self.remote_screen, name='remote')
        self.push_screen('remote')


if __name__ == '__main__':
    app = LolCatt('default')
    app.run()



from textual.containers import Container
from textual.reactive import reactive
from textual.widgets import Label
from textual.widgets import Static

from lolcatt.utils.utils import marquee


class LolCattDeviceInfo(Static):
    label_str = reactive('')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.label = Label('')
        self._marquee_gen = None

    def _get_device_info(self) -> str:
        info = self.app.caster.get_device_name()
        if info is not None:
            msg = f'Connected to: "{info}"'
        else:
            msg = 'Not connected to a device. Try "lolcatt --scan".'
        return msg

    def _update_label(self):
        self.label_str = self._get_device_info() + ' '
        self.label.update(next(self._marquee_gen))

    def watch_label_str(self, value):
        self._marquee_gen = marquee(value, self.size.width, 2)

    def on_resize(self, value):
        self._marquee_gen = marquee(self.label_str, self.size.width, 2)

    def on_mount(self):
        self._update_label()
        self.set_interval(
            interval=self.app.caster.get_update_interval(), callback=self._update_label
        )

    def compose(self):
        yield Container(self.label, id='device_info')



from dataclasses import dataclass

from catt.error import CastError
from textual import on
from textual.app import DEFAULT_COLORS
from textual.containers import Container
from textual.events import Key
from textual.widgets import Button
from textual.widgets import Static


@dataclass
class ControlsConfig:
    ffwd_secs: int = 30
    rewind_secs: int = 10
    vol_step: float = 0.1
    fancy_icons: bool = False


class LolCattControls(Static):
    CONTROLS = {
        'play_pause': '',
        'stop': '',
        'previous': '',
        'rewind': '',
        'ffwd': '',
        'next': '',
        'vol_down': '',
        'vol_up': '',
    }

    CONTROLS_ASCII = {
        'play_pause': 'Play/Pause',
        'stop': 'Stop',
        'previous': '|<',
        'rewind': '<<',
        'ffwd': '>>',
        'next': '>|',
        'vol_down': 'Vol-',
        'vol_up': 'Vol+',
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._config = ControlsConfig(
            fancy_icons=self.app.config['options'].get('fancy_icons', False)
        )
        self._pp_button_colors = [DEFAULT_COLORS['dark'].success, DEFAULT_COLORS['dark'].warning]
        self._pp_button = Button(self._get_control_label('play_pause'), id='play_pause_button')
        self._pp_button.styles.border_bottom = ('tall', self._pp_button_colors[0])
        self._stop_button = Button(self._get_control_label('stop'), id='stop_button')

    def _get_control_label(self, control: str) -> str:
        if self._config.fancy_icons:
            return self.CONTROLS[control]
        else:
            return self.CONTROLS_ASCII[control]

    def compose(self):
        with Container(id='controls'):
            with Container(id='playback_buttons'):
                yield self._pp_button
                yield self._stop_button

            with Container(id='wind_buttons'):
                yield Button(self._get_control_label('previous'), id='prev_button')
                yield Button(self._get_control_label('rewind'), id='rewind_button')
                yield Button(self._get_control_label('ffwd'), id='ffwd_button')
                yield Button(self._get_control_label('next'), id='next_button')

            with Container(id='volume_buttons'):
                yield Button(self._get_control_label('vol_down'), id='vol_down_button')
                yield Button(self._get_control_label('vol_up'), id='vol_up_button')

    @on(Button.Pressed, "#play_pause_button")
    def toggle_play_pause(self):
        try:
            self.app.caster.get_device().controller.play_toggle()
            self._pp_button.styles.border_bottom = ('tall', self._pp_button_colors[-1])
            self._pp_button_colors.reverse()
        except ValueError:
            pass

    @on(Button.Pressed, "#stop_button")
    def stop(self):
        if self.app.caster.get_cast_state().cast_info.get('player_state') in ['PLAYING', 'PAUSED']:
            self.app.caster.stop_cast()
        else:
            self.app.exit()

    @on(Button.Pressed, "#vol_down_button")
    def vol_down(self):
        self.app.caster.get_device().volumedown(self._config.vol_step)

    @on(Button.Pressed, "#vol_up_button")
    def vol_up(self):
        self.app.caster.get_device().volumeup(self._config.vol_step)

    @on(Button.Pressed, "#ffwd_button")
    def ffwd(self):
        try:
            self.app.caster.get_device().ffwd(self._config.ffwd_secs)
        except CastError:
            pass

    @on(Button.Pressed, "#rewind_button")
    def rewind(self):
        try:
            self.app.caster.get_device().rewind(self._config.rewind_secs)
        except CastError:
            pass

    @on(Button.Pressed, '#next_button')
    def next(self):
        try:
            self.app.caster.cast_next()
        except CastError:
            pass

    @on(Button.Pressed, '#prev_button')
    def prev(self):
        try:
            self.app.caster.cast_previous()
        except CastError:
            pass

    def on_key(self, event: Key) -> None:
        if event.key == 'space':
            self.toggle_play_pause()
        elif event.key == ('q'):
            self.app.exit()
        elif event.key in ('h', 'left'):
            self.rewind()
        elif event.key in ('l', 'right'):
            self.ffwd()
        elif event.key in ('j', 'down'):
            self.next()
        elif event.key in ('k', 'up'):
            self.prev()
        elif event.key == 'plus':
            self.vol_up()
        elif event.key == 'minus':
            self.vol_down()
        else:
            pass



from typing import Tuple

from catt.error import CastError
from textual.containers import Container
from textual.events import Click
from textual.reactive import reactive
from textual.widgets import Label
from textual.widgets import ProgressBar
from textual.widgets import Static


class LolCattProgress(Static):
    current = reactive(0)
    duration = reactive(0)
    percent_complete = reactive(0)

    @staticmethod
    def _extract_progress(cast_info: dict) -> Tuple[float, float, float]:
        current = cast_info.get('current_time', 0.0)
        duration = cast_info.get('duration', 0.0)
        percent_complete = current / duration * 100 if duration else 0.0
        return current, duration, percent_complete

    def _format_time(self, seconds: float) -> str:
        if seconds is None:
            return '--:--'
        minutes, seconds = divmod(seconds, 60)
        hours, minutes = divmod(minutes, 60)
        if hours:
            return f'{hours:02.0f}:{minutes:02.0f}:{seconds:02.0f}'
        return f'{minutes:02.0f}:{seconds:02.0f}'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pb = ProgressBar(total=100, show_percentage=False, show_eta=False, id='progress_bar')
        self.pblabel = Label('', id='progress_label')
        self._needed_width = 16  # FIXME: learn CSS

    def update_progress(self) -> int:
        self.current, self.duration, self.percent_complete = self._extract_progress(
            self.app.caster.get_cast_state().cast_info
        )
        self.pb.update(progress=self.percent_complete)
        current_fmt = self._format_time(self.current)
        duration_fmt = self._format_time(self.duration)
        upd_str = f'({current_fmt}/{duration_fmt})'
        padding = ' ' * max(0, (self._needed_width - len(upd_str)))
        self.pblabel.update(f'{padding}{upd_str}')

    def on_mount(self):
        self.update_progress()
        self.set_interval(
            interval=self.app.caster.get_update_interval(), callback=self.update_progress
        )

    def on_click(self, event: Click):
        min_x, max_x = (
            self.pb.content_region.x,
            self.pb.content_region.x + self.pb.content_region.width,
        )
        click_x = min(max(event.screen_x, min_x), max_x)
        fraction = min(1, (click_x - min_x) / (max_x - min_x))
        duration = self.app.caster.get_cast_state().cast_info.get('duration', 0.0)
        try:
            self.app.caster.get_device().seek(duration * fraction)
        except CastError:
            pass

    def compose(self):
        yield Container(self.pb, self.pblabel, id='progress')



from textual import on
from textual.containers import Container
from textual.events import Key
from textual.widgets import Input
from textual.widgets import Static


class LolCattUrlInput(Static):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._input = Input(id='url_input', placeholder='Enter URL to cast...')
        self._input.cursor_blink = False

    @on(Input.Submitted, "#url_input")
    def cast_url(self):
        if self._input.value == '':
            return
        if self._input.value:
            self.app.caster.enqueue(self._input.value, front=True)
            print(self.app.caster._queue)
            self.app.caster.cast_next()
            self._input.value = ''

    def compose(self):
        yield Container(self._input, id='url_input_container')

    def on_key(self, event: Key):
        if event.key == 'escape':
            self.app.remote_screen.focus_next()

    def on_mount(self):
        self._input.focus()



#!/usr/bin/env python3
from pathlib import Path

from catt.api import discover


def marquee(s: str, max_len: int, step_size: int) -> str:
    """
    Yields a string that is max_len characters long, and shifts the string by step_size each time.

    :param s: The string to be shifted
    :param max_len: The maximum length of the string
    :param step_size: The number of characters to shift the string by
    :return: A string that is max_len characters long and appropriately shifted
    """
    while True:
        if len(s) < max_len:
            yield s
        else:
            s = f'{s[step_size:]}{s[:step_size]}'
            yield s[:max_len]


def scan():
    """Scans the network for Chromecast devices and prints the results."""
    print('Scanning for Chromecast devices...')

    discovered = discover()
    deduped = []
    found_ips = set()
    max_name_len = 11  # len('Device name')
    for device in discovered:
        if device.ip_addr not in found_ips:
            found_ips.add(device.ip_addr)
            max_name_len = max(max_name_len, len(device.name))
            deduped.append((device.name, device.ip_addr))

    deduped = [('"' + x + '"' + ' ' * (max_name_len - len(x)), y) for x, y in deduped]
    deduped = sorted(deduped, key=lambda x: x[1])

    print('Found {} device(s):'.format(len(found_ips)))
    print(f'Device name{" " * (max_name_len - 11)}\tDevice IP')
    print('=' * (max_name_len + 19))
    for name, ip in deduped:
        print(f'{name}\t{ip}')


def write_initial_config(p: Path):
    p.write_text(
        "[options]\n"
        "fancy_icons = true  # Whether to use fancy icons\n"
        "#youtube_cookies_file = \"~/.config/lolcatt/cookies.txt\"  # Path to a cookies.txt file for YouTube"
    )



from textual.containers import Container
from textual.reactive import reactive
from textual.widgets import Label
from textual.widgets import Static

from lolcatt.utils.utils import marquee


class LolCattPlaybackInfo(Static):
    label_str = reactive('')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.label = Label('', id='title')
        self._marquee_gen = None

    def _get_playback_info(self) -> str:
        playing = self.app.caster.get_cast_state().cast_info.get('title')
        display_name = self.app.caster.get_cast_state().info.get('display_name')
        is_loading = self.app.caster.get_cast_state().is_loading

        if playing is not None:
            return f'Playing: "{playing}"'
        elif display_name is not None and display_name != 'Backdrop':
            return f'Displaying: "{display_name}"'
        elif is_loading:
            return 'Loading...'
        else:
            return 'Nothing is playing.'

    def _update_label(self):
        self.label_str = self._get_playback_info() + ' '
        self.label.update(next(self._marquee_gen))

    def watch_label_str(self, value):
        self._marquee_gen = marquee(value, self.size.width, 2)

    def on_resize(self, value):
        self._marquee_gen = marquee(self.label_str, self.size.width, 2)

    def compose(self):
        yield Container(self.label, id='playback_info')

    def on_mount(self):
        self.set_interval(
            interval=self.app.caster.get_update_interval(), callback=self._update_label
        )



#!/usr/bin/env python3
from pathlib import Path

import click
import toml

from lolcatt.app import LolCatt
from lolcatt.utils.utils import scan as do_scan
from lolcatt.utils.utils import write_initial_config


@click.command(
    'lolcatt',
    context_settings=dict(help_option_names=['-h', '--help']),
)
@click.argument(
    'url_or_path',
    nargs=1,
    default=None,
    required=False,
)
@click.option(
    '-d',
    '--device',
    default='default',
    help='Device name or alias (defined in catt config) to cast to. '
    'Per default, uses the device noted as default in the `catt` config file. '
    'If no default is set, the first device found will be used.',
)
@click.option(
    '--scan',
    is_flag=True,
    default=False,
    help='Scan for Chromecast devices and exit, printing found devices.',
)
@click.option(
    '--config',
    default='~/.config/lolcatt/config.toml',
    type=click.Path(dir_okay=False),
    help='Path to catt config file. Defaults to ~/.config/lolcatt/config.toml -'
    'If the file does not exist, it will be created.',
)
def main(url_or_path, device, scan, config):
    """Cast media from a local file or URL to a Chromecast device."""
    config = Path(str(config)).expanduser()
    if not config.exists():
        config.parent.mkdir(parents=True, exist_ok=True)
        write_initial_config(config)

    config = toml.loads(config.read_text())

    if url_or_path is None and scan:
        do_scan()
        return

    lolcatt = LolCatt(device_name=device, config=config)
    if url_or_path is not None:
        lolcatt.caster.enqueue(url_or_path)
        lolcatt.caster.cast_next()
    lolcatt.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""Tests for `lolcatt` package."""
from lolcatt.app import LolCatt


def test_lolcat_start():
    """Simple smoke test for now."""
    lolcatt = LolCatt(device_name=None)
    lolcatt.run()



#!/usr/bin/env python
"""The setup script."""
from pkg_resources import parse_requirements
from setuptools import find_packages
from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

history = ''

with open('requirements/prod.txt') as prod_req:
    requirements = [str(ir) for ir in parse_requirements(prod_req)]
with open('requirements/test.txt') as test_req:
    test_requirements = [str(ir) for ir in parse_requirements(test_req)]

setup(
    author="Lukas Lüftinger",
    author_email='lukas.lueftinger@outlook.com',
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    description="A TUI wrapper around `catt`.",
    entry_points={
        'console_scripts': [
            'lolcatt=lolcatt.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='lolcatt',
    name='lolcatt',
    packages=find_packages(include=['lolcatt', 'lolcatt.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/LokiLuciferase/lolcatt',
    version='0.2.0',
    zip_safe=False,
)



#!/usr/bin/env python3
import subprocess
import time
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional

from catt.api import CattDevice
from catt.api import discover
from catt.cli import get_config_as_dict
from catt.error import CastError

from .youtube_playlist_handler import YoutubePlaylistHandler


@dataclass
class CastState:
    """Dataclass for cast state, encapsulating info dictionaries of a catt controller."""

    cast_info: dict
    info: dict
    is_loading: bool = False
    media_loaded: bool = False


class Caster:
    """
    Class encapsulating the catt.api.CattDevice.

    Provides a simple interface and enables exchange of the CattDevice on the fly.
    """

    CATT_ARGS = []
    CAST_ARGS = ['--force-default']

    def __init__(
        self,
        name_or_alias: Optional[str] = 'default',
        update_interval: float = 0.5,
        autoplay: bool = True,
        config: Dict = None,
    ):
        self._config = config if config is not None else {}
        self._yt_playlist_handler = YoutubePlaylistHandler(
            cookies_file=self._config.get('options', {}).get('youtube_cookies_file')
        )
        self._device = None
        self._queue = []
        self._current_item = None
        self._played_queue = []
        self._available_devices = None
        self._catt_call = None
        self._catt_config = get_config_as_dict()
        self._loading_started = time.time()
        self._is_loading_cast = True
        self._media_loaded = False
        self._loading_timeout = 8
        self._update_interval = update_interval
        self._autoplay = autoplay
        self._state_last_updated = time.time()
        self.change_device(name_or_alias)

    def cast(self, url_or_path: str):
        """
        Casts the given url or path to the currently active device.

        :param url_or_path: The url or path to cast.
        """
        if self._catt_call is not None:
            self._catt_call.kill()
        if self._device is None:
            raise ValueError('Can\'t cast: No device selected.')
        full_catt_args = [
            'catt',
            *self.CATT_ARGS,
            '-d',
            self._device_name,
            'cast',
            *self.CAST_ARGS,
            url_or_path,
        ]
        self._catt_call = subprocess.Popen(
            full_catt_args,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        self._trigger_is_loading()

    def cast_next(self):
        """
        Casts the next item in the queue to the currently active device.
        """
        if self._current_item is not None:
            self._played_queue.append(self._current_item)
            self._current_item = None
        if len(self._queue) > 0:
            self._current_item = self._queue.pop(0)
            self.cast(self._current_item)
        else:
            try:
                self._device.stop()
            except CastError:
                pass

    def cast_previous(self):
        """
        Casts the previous item in the queue to the currently active device.
        """
        if self._current_item is not None:
            self._queue.insert(0, self._current_item)
            self._current_item = None
        if len(self._played_queue) > 0:
            self._current_item = self._played_queue.pop()
            self.cast(self._current_item)

    def stop_cast(self):
        """
        Stops the current cast.
        """
        if self._current_item is not None:
            self._played_queue.append(self._current_item)
            self._current_item = None
        if self._catt_call is not None:
            self._catt_call.kill()
        if self._device is not None:
            self._device.stop()

    def enqueue(self, url_or_path: str, front: bool = False):
        """
        Enqueues the given url or path to the queue.

        :param url_or_path: The url or path to enqueue.
        :param front: If True, the url or path is added to the front of the queue.
        """
        if 'youtube' in url_or_path and 'playlist?' in url_or_path:
            self._trigger_is_loading()
            all_new_urls = self._yt_playlist_handler.resolve_playlist(url_or_path)
            self._queue = all_new_urls + self._queue if front else self._queue + all_new_urls
        else:
            self._queue.append(url_or_path) if not front else self._queue.insert(0, url_or_path)

    def clear_queue(self):
        """
        Clears the queue.
        """
        self._queue = []

    def get_queue(self) -> List[str]:
        """
        Returns the current queue.

        :return: The current queue.
        """
        return self._queue

    def get_available_devices(self) -> List[str]:
        """
        Runs Chromecast discovery and returns a list of available CattDevices.

        :return: A list of available CattDevices.
        """
        self._available_devices = discover()
        return self._available_devices

    def change_device(self, name_or_alias: str = None):
        """
        Changes the currently active device to the given name or alias. If the device is not
        available, a ValueError is raised.

        :param name_or_alias: The name or alias of the device to change to.
        """
        if name_or_alias == 'default':
            self._device_name = self._catt_config['options'].get('device')
            if self._device_name is None:
                print(
                    'No default device set. '
                    'Scanning for all available devices and picking first...'
                )
                print(
                    'To skip this in the future, either pass a device name '
                    'or set a default device in the catt config file.'
                )
                possible_devices = self.get_available_devices()
                if len(possible_devices) > 0:
                    self._device_name = possible_devices[0].name
        elif name_or_alias == 'None':
            self._device_name = None
        elif name_or_alias is not None:
            self._device_name = self._catt_config['aliases'].get(name_or_alias, name_or_alias)
        else:
            self._device_name = None

        if self._device_name is not None:
            try:
                self._device = CattDevice(self._device_name)
                self._loading_started = None
                self._is_loading_cast = False
            except CastError:
                print(f'Selected device "{self._device_name}" was not found on this network.')
                print('Scan for available devices using "lolcatt --scan".')
                raise ValueError(f'No device with name or alias "{name_or_alias}" found.')

    def get_device(self) -> CattDevice:
        """
        Returns the currently active CattDevice.

        :return: The currently active CattDevice.
        """
        return self._device

    def get_device_name(self) -> Optional[str]:
        """
        Returns the name of the currently active CattDevice.

        :return: The name of the currently active CattDevice.
        """
        return self._device_name

    def get_update_interval(self) -> float:
        """
        Returns the update interval of the CastState. Determines how often UI elements are need to
        be updated.

        :return: The update interval of the CastState.
        """
        return self._update_interval

    def _trigger_is_loading(self):
        self._loading_started = time.time()
        self._is_loading_cast = True

    def _tick(self):
        """
        Internal method that is called on every call to update the CastState.
        """
        if (
            self._autoplay
            and self._device.controller.info.get('idle_reason') == 'FINISHED'
            and len(self._queue) > 0
        ):
            self.cast_next()

        if self._device is None:
            return

        if time.time() - self._state_last_updated > self._update_interval:
            self._device.controller._update_status()
            self._state_last_updated = time.time()

        if self._is_loading_cast and time.time() - self._loading_started > self._loading_timeout:
            self._loading_started = None
            self._is_loading_cast = False

        if self._device.controller.cast_info.get('player_state') in [
            'PLAYING',
            'BUFFERING',
            'PAUSED',
        ]:
            self._loading_started = None
            self._is_loading_cast = False
            self._media_loaded = True
        else:
            self._media_loaded = False

    def get_cast_state(self) -> CastState:
        """
        Returns a CastState object encapsulating the info dictionaries of the currently active
        CattDevice.

        :return: A CastState object
        """
        self._tick()
        if self._device is None:
            cs = CastState({}, {}, False, False)
        else:
            cs = CastState(
                self._device.controller.cast_info,
                self._device.controller.info,
                is_loading=self._is_loading_cast,
                media_loaded=self._media_loaded,
            )
        return cs
